/*
**  NSRegEx.h
**
**  Copyright (c) 2001, 2002
**
**  Author: Francis Lachapelle <francis@Sophos.ca>
**          Ludovic Marcotte <ludovic@Sophos.ca>
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**  
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
**  Lesser General Public License for more details.
**  
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#ifndef _Pantomime_H_NSRegEx
#define _Pantomime_H_NSRegEx

#include <sys/types.h>
#include <regex.h>

#import <Foundation/NSObject.h>
#import <Foundation/NSString.h>

@interface NSRegEx : NSObject
{
  regex_t re;
}

//
// flags defaults to REG_EXTENDED in initializers not taking flags
//
- (id) initWithPattern: (NSString *) thePattern;
- (id) initWithPattern: (NSString *) thePattern
                 flags: (int) theFlags;

+ (id) regexWithPattern: (NSString *) thePattern;
+ (id) regexWithPattern: (NSString *) thePattern
                  flags: (int) theFlags;

- (NSArray *) matchString: (NSString *) theString;

+ (NSArray *) matchString: (NSString *) theString
              withPattern: (NSString *) pattern
          isCaseSensitive: (BOOL) caseSensitive;

@end

#endif // _Pantomime_H_NSRegEx
