/*
**  LocalFolderCacheManager.h
**
**  Copyright (c) 2001, 2002
**
**  Author: Ludovic Marcotte <ludovic@Sophos.ca>
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**  
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
**  Lesser General Public License for more details.
**  
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#ifndef _Pantomime_H_LocalFolderCacheManager
#define _Pantomime_H_LocalFolderCacheManager

#import <Foundation/NSArray.h>
#import <Foundation/NSCoder.h>
#import <Foundation/NSDate.h>
#import <Foundation/NSObject.h>
#import <Foundation/NSString.h>

@class LocalMessage;

@interface LocalFolderCacheManager: NSObject <NSCoding>
{
  NSMutableArray *messages;

  NSString *pathToCache;
  NSString *pathToFolder;

  NSDate *modificationDate;
  int fileSize;

  BOOL dirty;
}

- (id) init;
- (id) initWithPath: (NSString *) thePath;
- (void) dealloc;

- (BOOL) synchronize;

//
// NSCoding protocol
//

- (void) encodeWithCoder: (NSCoder *) theCoder;
- (id) initWithCoder: (NSCoder *) theCoder;

//
// access/mutation methods
//

- (NSDate *) modificationDate;
- (void) setModificationDate: (NSDate *) theDate;

- (int) fileSize;
- (void) setFileSize: (int) theSize;

- (void) addMessage: (LocalMessage *) theMessage;
- (void) removeMessage: (LocalMessage *) theMessage;

- (NSArray *) messages;
- (void) setMessages: (NSArray *) theMessages;

- (NSString *) pathToCache;
- (void) setPathToCache: (NSString *) thePathToCache;

- (NSString *) pathToFolder;
- (void) setPathToFolder: (NSString *) thePathToFolder;

//
// Other methods
//

- (void) invalidate;

//
// class methods
//

+ (id) localFolderCacheFromDiskWithPath: (NSString *) thePath;

@end

#endif // _Pantomime_H_LocalFolderCacheManager
