/*
**  FolderInformation.m
**
**  Copyright (c) 2002
**
**  Author: Ludovic Marcotte <ludovic@Sophos.ca>
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**  
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
**  Lesser General Public License for more details.
**  
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#include <Pantomime/FolderInformation.h>

#include <Pantomime/Constants.h>

@implementation FolderInformation

- (id) init
{
  self = [super init];

  [self setNbOfMessages: 0];
  [self setNbOfUnreadMessages: 0];
  [self setSize: 0];

  return self;
}

- (void) dealloc
{
  [super dealloc];
}

//
// access/mutation methods
//
- (int) nbOfMessages
{
  return nbOfMessages;
}

- (void) setNbOfMessages: (int) theValue
{
  nbOfMessages = theValue;
}

- (int) nbOfUnreadMessages
{
  return nbOfUnreadMessages;
}

- (void) setNbOfUnreadMessages: (int) theValue
{
  nbOfUnreadMessages = theValue;
}

- (int) size
{
  return size;
}

- (void) setSize: (int) theSize
{
  size = theSize;
}

@end
